"""
# TODO: Update description
Tekton pipeline for with the following steps:
1. Step 1
"""

# TODO: Install kfp
# pip install kfp
# pip install kfp-tekton

import kfp
import kfp.dsl as dsl
import kfp.components as comp
from kfp_tekton.compiler import TektonCompiler

# TODO: Add your pipeline components based on the kfp yaml file from CLAIMED
# initialize operator from yaml file
component_op = comp.load_component_from_file('<operator>.yaml')
# initialize operator from remote file
web_op = comp.load_component_from_url('https://raw.githubusercontent.com/claimed-framework/component-library/main/component-library/<operator>.yaml')


# TODO: Update pipeline description, function name, and parameters
pipeline_name = 'my_pipeline'
# Pipeline function
@dsl.pipeline(
    name=pipeline_name,
    description="Pipeline description"
)
def my_pipeline(
    parameter1: str = "default_value",
    parameter2: str = "default_value",
):
    # TODO: Add the components and the required parameters
    step1 = component_op(
        parameter1=parameter1,
    )
    step2 = web_op(
        parameter2=parameter2,
    )

    # TODO: You can call multiple steps and created the dependencies
    step2.after(step1)

# TODO: Update pipeline function
# Kubernetes
kfp.compiler.Compiler().compile(pipeline_func=my_pipeline, package_path=f'{pipeline_name}.yaml')
# OpenShift with Tekton
TektonCompiler().compile(my_pipeline, f'{pipeline_name}.yaml')

print(f'Saved pipeline in {pipeline_name}.yaml')

# TODO: Run script with python
# TODO: Upload the yaml to KubeFlow




import os
import sys
import logging
import shutil
import argparse
import subprocess
from string import Template
from importlib.metadata import version
from c3.pythonscript import Pythonscript
from c3.utils import convert_notebook, get_image_version
from c3.templates import component_setup_code, dockerfile_template, kfp_component_template, kubernetes_job_template

CLAIMED_VERSION = 'V0.1'


def create_operator(file_path: str,
                    repository: str,
                    version: str,
                    dockerfile_template: str,
                    additional_files: str = None,
                    log_level='INFO',
                    ):
    logging.info('Parameters: ')
    logging.info('file_path: ' + file_path)
    logging.info('repository: ' + repository)
    logging.info('version: ' + str(version))
    logging.info('additional_files: ' + str(additional_files))

    # TODO: add argument for running ipython instead of python within the container
    if file_path.endswith('.ipynb'):
        logging.info('Convert notebook to python script')
        target_code = convert_notebook(file_path)
    elif file_path.endswith('.py'):
        target_code = file_path.split('/')[-1]
        if file_path == target_code:
            # use temp file for processing
            target_code = 'claimed_' + target_code
        # Copy file to current working directory
        shutil.copy(file_path, target_code)
    else:
        raise NotImplementedError('Please provide a file_path to a jupyter notebook or python script.')

    if target_code.endswith('.py'):
        # Add code for logging and cli parameters to the beginning of the script
        with open(target_code, 'r') as f:
            script = f.read()
        script = component_setup_code + script
        with open(target_code, 'w') as f:
            f.write(script)

        # getting parameter from the script
        py = Pythonscript(target_code)
        name = py.get_name()
        # convert description into a string with a single line
        description = ('"' + py.get_description().replace('\n', ' ').replace('"', '\'') +
                       ' – CLAIMED ' + CLAIMED_VERSION + '"')
        inputs = py.get_inputs()
        outputs = py.get_outputs()
        requirements = py.get_requirements()
    else:
        raise NotImplementedError('C3 currently only supports jupyter notebook or python script.')
    
    # Strip 'claimed-' from name of copied temp file
    if name.startswith('claimed-'):
        name = name[8:]

    logging.info('Operator name: ' + name)
    logging.info('Description:: ' + description)
    logging.info('Inputs: ' + str(inputs))
    logging.info('Outputs: ' + str(outputs))
    logging.info('Requirements: ' + str(requirements))

    # copy all additional files to temporary folder
    additional_files_path = 'additional_files_path'
    while os.path.exists(additional_files_path):
        # ensures using a new directory
        additional_files_path += '_temp'
    logging.debug(f'Create dir for additional files {additional_files_path}')
    os.makedirs(additional_files_path)
    for additional_file in additional_files:
        assert os.path.isfile(additional_file), \
            f"Could not find file at {additional_file}. Please provide only files as additional parameters."
        shutil.copy(additional_file, additional_files_path)
    logging.info(f'Selected additional files: {os.listdir(additional_files_path)}')

    requirements_docker = list(map(lambda s: 'RUN ' + s, requirements))
    requirements_docker = '\n'.join(requirements_docker)

    docker_file = dockerfile_template.substitute(
        requirements_docker=requirements_docker,
        target_code=target_code,
        additional_files_path=additional_files_path,
    )

    logging.info('Create Dockerfile')
    with open("Dockerfile", "w") as text_file:
        text_file.write(docker_file)
        
    if version is None:
        # auto increase version based on registered images
        version = get_image_version(repository, name)

    logging.info(f'Building container image claimed-{name}:{version}')
    try:
        subprocess.run(
            ['docker', 'build', '--platform', 'linux/amd64', '-t', f'claimed-{name}:{version}', '.'],
            stdout=None if log_level == 'DEBUG' else subprocess.PIPE, check=True,
        )
        logging.debug(f'Tagging images with "latest" and "{version}"')
        subprocess.run(
            ['docker', 'tag', f'claimed-{name}:{version}', f'{repository}/claimed-{name}:{version}'],
            stdout=None if log_level == 'DEBUG' else subprocess.PIPE, check=True,
        )
        subprocess.run(
            ['docker', 'tag', f'claimed-{name}:{version}', f'{repository}/claimed-{name}:latest'],
            stdout=None if log_level == 'DEBUG' else subprocess.PIPE, check=True,
        )
        logging.info('Successfully built image')
    except:
        logging.error(f'Failed to build image with docker.')
        pass

    logging.info(f'Pushing images to registry {repository}')
    try:
        subprocess.run(
            ['docker', 'push', f'{repository}/claimed-{name}:latest'],
            stdout=None if log_level == 'DEBUG' else subprocess.PIPE, check=True,
        )
        subprocess.run(
            ['docker', 'push', f'{repository}/claimed-{name}:{version}'],
            stdout=None if log_level == 'DEBUG' else subprocess.PIPE, check=True,
        )
        logging.info('Successfully pushed image to registry')
    except:
        logging.error(f'Could not push images to namespace {repository}. '
                      f'Please check if docker is logged in or select a namespace with access.')
        pass

    def get_component_interface(parameters):
        return_string = str()
        for name, options in parameters.items():
            return_string += f'- {{name: {name}, type: {options["type"]}, description: "{options["description"]}"'
            if options['default'] is not None:
                if not options["default"].startswith("'"):
                    options["default"] = f"'{options['default']}'"
                return_string += f', default: {options["default"]}'
            return_string += '}\n'
        return return_string
    inputs_list = get_component_interface(inputs)
    outputs_list = get_component_interface(outputs)

    parameter_list = str()
    for index, key in enumerate(list(inputs.keys()) + list(outputs.keys())):
        parameter_list += f'{key}="${{{index}}}" '

    parameter_values = str()
    for input_key in inputs.keys():
        parameter_values += f"        - {{inputValue: {input_key}}}\n"
    for input_key in outputs.keys():
        parameter_values += f"        - {{outputPath: {input_key}}}\n"

    yaml = kfp_component_template.substitute(
        name=name,
        description=description,
        repository=repository,
        version=version,
        inputs=inputs_list,
        outputs=outputs_list,
        call=f'./{target_code} {parameter_list}',
        parameter_values=parameter_values,
    )

    logging.debug('KubeFlow component yaml:\n' + yaml)
    target_yaml_path = file_path.replace('.ipynb', '.yaml').replace('.py', '.yaml')

    logging.info(f'Write KubeFlow component yaml to {target_yaml_path}')
    with open(target_yaml_path, "w") as text_file:
        text_file.write(yaml)

    # get environment entries
    env_entries = str()
    for key in list(inputs.keys()) + list(outputs.keys()):
        env_entries += f"        - name: {key}\n          value: value_of_{key}\n"
    env_entries = env_entries.rstrip()

    job_yaml = kubernetes_job_template.substitute(
        name=name,
        repository=repository,
        version=version,
        target_code=target_code,
        env_entries=env_entries,
    )

    logging.debug('Kubernetes job yaml:\n' + job_yaml)
    target_job_yaml_path = file_path.replace('.ipynb', '.job.yaml').replace('.py', '.job.yaml')

    logging.info(f'Write kubernetes job yaml to {target_job_yaml_path}')
    with open(target_job_yaml_path, "w") as text_file:
        text_file.write(job_yaml)

    logging.info(f'Remove local files')
    # remove temporary files
    if file_path != target_code:
        os.remove(target_code)
    os.remove('Dockerfile')
    if additional_files_path is not None:
        shutil.rmtree(additional_files_path, ignore_errors=True)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('FILE_PATH', type=str,
                        help='Path to python script or notebook')
    parser.add_argument('ADDITIONAL_FILES', type=str, nargs='*',
                        help='Paths to additional files to include in the container image')
    parser.add_argument('-r', '--repository', type=str, required=True,
                        help='Container registry address, e.g. docker.io/<your_username>')
    parser.add_argument('-v', '--version', type=str, default=None,
                        help='Image version. Increases the version numer of image:latest if not provided.')
    parser.add_argument('-l', '--log_level', type=str, default='INFO')
    parser.add_argument("-v", "--version", action="store_true")
    parser.add_argument('--dockerfile_template_path', type=str, default='',
                        help='Path to custom dockerfile template')
    args = parser.parse_args()

    if args.version:
        print(version("claimed-c3"))
        sys.exit()

    # Init logging
    root = logging.getLogger()
    root.setLevel(args.log_level)
    handler = logging.StreamHandler(sys.stdout)
    formatter = logging.Formatter('%(levelname)s - %(message)s')
    handler.setFormatter(formatter)
    handler.setLevel(args.log_level)
    root.addHandler(handler)

    # Update dockerfile template if specified
    if args.dockerfile_template_path != '':
        logging.info(f'Uses custom dockerfile template from {args.dockerfile_template_path}')
        with open(args.dockerfile_template_path, 'r') as f:
            _dockerfile_template = Template(f.read())
    else:
        _dockerfile_template = dockerfile_template

    create_operator(
        file_path=args.FILE_PATH,
        repository=args.repository,
        version=args.version,
        dockerfile_template=_dockerfile_template,
        additional_files=args.ADDITIONAL_FILES,
        log_level=args.log_level,
    )


if __name__ == '__main__':
    main()



import os
import logging
import json
import re
import subprocess


def convert_notebook(path):
    # TODO: switch to nbconvert long-term (need to replace pip install)
    with open(path) as json_file:
        notebook = json.load(json_file)

    # backwards compatibility
    if notebook['cells'][0]['cell_type'] == 'markdown' and notebook['cells'][1]['cell_type'] == 'markdown':
        logging.info('Merge first two markdown cells. File name is used as operator name, not first markdown cell.')
        notebook['cells'][1]['source'] = notebook['cells'][0]['source'] + ['\n'] + notebook['cells'][1]['source']
        notebook['cells'] = notebook['cells'][1:]

    code_lines = []
    for cell in notebook['cells']:
        if cell['cell_type'] == 'markdown':
            # add markdown as doc string
            code_lines.extend(['"""\n'] + [f'{line}' for line in cell['source']] + ['\n"""'])
        elif cell['cell_type'] == 'code' and cell['source'][0].startswith('%%bash'):
            code_lines.append('os.system("""')
            code_lines.extend(cell['source'][1:])
            code_lines.append('""")')
        elif cell['cell_type'] == 'code':
            for line in cell['source']:
                if line.strip().startswith('!'):
                    # convert sh scripts
                    if re.search('![ ]*pip', line):
                        # change pip install to comment
                        code_lines.append(re.sub('![ ]*pip', '# pip', line))
                    else:
                        # change sh command to os.system()
                        logging.info(f'Replace shell command with os.system() ({line})')
                        code_lines.append(line.replace('!', 'os.system(', 1).replace('\n', ')\n'))
                else:
                    # add code
                    code_lines.append(line)
        # add line break after cell
        code_lines.append('\n')
    code = ''.join(code_lines)

    py_path = path.split('/')[-1].replace('.ipynb', '.py')

    assert not os.path.exists(py_path), f"File {py_path} already exist. Cannot convert notebook."
    with open(py_path, 'w') as py_file:
        py_file.write(code)

    return py_path


def increase_image_version(last_version):
    try:
        # increase last version value by 1
        version = last_version.split('.')
        version[-1] = str(int(version[-1]) + 1)
        version = '.'.join(version)
    except:
        # fails if a string value was used for the last tag
        version = last_version + '.1'
        logging.debug(f'Failed to increase last value, adding .1')
        pass
    logging.info(f'Using version {version} based on latest tag ({last_version}).')
    return version


def pull_docker_image_tags(image):
    logging.warning("The current implementation can only query local docker images. "
                    "Please use an argument '-v <version>' to avoid duplicates.")
    # list images
    output = subprocess.run(
        ['docker', 'image', 'ls', image],
        stdout=subprocess.PIPE
    ).stdout.decode('utf-8')
    try:
        # remove header
        image_list = output.splitlines()[1:]
        # get list of image tags
        image_tags = [line.split()[1] for line in image_list]
    except:
        image_tags = []
        logging.error(f"Could not load image tags from 'docker image ls' output: {output}")
        pass

    # filter latest and none
    image_tags = [t for t in image_tags if t not in ['latest', '<none>']]
    return image_tags


def pull_icr_image_tags(image):
    # list images from icr
    output = subprocess.run(
        ['ibmcloud', 'cr', 'images', '--restrict', image.split('icr.io/', 1)[1]],
        stdout=subprocess.PIPE
    ).stdout.decode('utf-8')

    try:
        # remove header and final status
        image_list = output.splitlines()[3:-2]
        # get list of image tags
        image_tags = [line.split()[1] for line in image_list]
    except:
        image_tags = []
        logging.error(f"Could not load image tags from 'ibmcloud cr images' output: {output}")
        pass

    # filter latest and none
    image_tags = [t for t in image_tags if t not in ['latest', '<none>']]
    return image_tags


def get_image_version(repository, name):
    """
    Get current version of the image from the registry and increase the version by 1.
    Defaults to 0.1 if no image is found in the registry.
    """
    logging.debug(f'Get image version from registry.')
    if 'docker.io' in repository:
        logging.debug('Get image tags from docker.')
        image_tags = pull_docker_image_tags(f'{repository}/claimed-{name}')
    elif 'icr.io' in repository:
        logging.debug('Get image tags from ibmcloud container registry.')
        image_tags = pull_icr_image_tags(f'{repository}/claimed-{name}')
    else:
        logging.warning('Unrecognised container registry, using docker to query image tags.')
        image_tags = pull_docker_image_tags(f'{repository}/claimed-{name}')
    logging.debug(f'Image tags: {image_tags}')

    def check_only_numbers(test_str):
        return set(test_str) <= set('.0123456789')

    if len(image_tags) == 0:
        # default version
        version = '0.1'
        logging.info(f'Using default version {version}. No prior image tag found for {repository}/claimed-{name}.')

    elif not check_only_numbers(image_tags[0]):
        # increase last version
        version = increase_image_version(image_tags[0])
        logging.info(f'Using version {version} based on last version {image_tags[0]}.')

    else:
        # find the highest numerical version
        image_tags = list(filter(check_only_numbers, image_tags))
        image_tags.sort(key=lambda s: list(map(int, s.split('.'))))
        version = increase_image_version(image_tags[-1])
        logging.info(f'Using version {version} based on highest previous version {image_tags[-1]}.')

    return version




import logging
import os
import argparse
import sys
from string import Template
from importlib.metadata import version
from c3.pythonscript import Pythonscript
from c3.utils import convert_notebook
from c3.create_operator import create_operator
from c3.templates import grid_wrapper_template, cos_grid_wrapper_template, gw_component_setup_code, dockerfile_template


def wrap_component(component_path,
                   component_description,
                   component_dependencies,
                   component_interface,
                   component_inputs,
                   component_process,
                   cos,
                   ):
    # get component name from path
    component_name = os.path.splitext(os.path.basename(component_path))[0]

    gw_template = cos_grid_wrapper_template if cos else grid_wrapper_template
    grid_wrapper_code = gw_template.substitute(
        component_name=component_name,
        component_description=component_description,
        component_dependencies=component_dependencies,
        component_inputs=component_inputs,
        component_interface=component_interface,
        component_process=component_process,
    )

    # Write edited code to file
    grid_wrapper_file = f'cgw_{component_name}.py' if cos else f'gw_{component_name}.py'
    grid_wrapper_file_path = os.path.join(os.path.dirname(component_path), grid_wrapper_file)
    # remove 'component_' from gw path
    grid_wrapper_file_path = grid_wrapper_file_path.replace('component_', '')
    with open(grid_wrapper_file_path, 'w') as f:
        f.write(grid_wrapper_code)

    logging.info(f'Saved wrapped component to {grid_wrapper_file_path}')

    return grid_wrapper_file_path


def get_component_elements(file_path):
    # get required elements from component code
    py = Pythonscript(file_path)
    # convert description into a string with a single line
    description = (py.get_description().replace('\n', ' ').replace('"', '\''))
    inputs = py.get_inputs()
    outputs = py.get_outputs()
    dependencies = py.get_requirements()

    # combine inputs and outputs
    interface_values = {}
    interface_values.update(inputs)
    interface_values.update(outputs)

    # combine dependencies list
    dependencies = '\n# '.join(dependencies)

    # generate interface code from inputs and outputs
    interface = ''
    type_to_func = {'String': '', 'Boolean': 'bool', 'Integer': 'int', 'Float': 'float'}
    for variable, d in interface_values.items():
        interface += f"# {d['description']}\n"
        interface += f"component_{variable} = {type_to_func[d['type']]}(os.getenv('{variable}', {d['default']}))\n"

    # generate kwargs for the subprocesses
    process_inputs = ', '.join([f'{i}=component_{i}' for i in inputs.keys()])
    # use log level from grid wrapper
    process_inputs = process_inputs.replace('component_log_level', 'log_level')

    return description, interface, process_inputs, dependencies


# Adding code
def edit_component_code(file_path):
    file_name = os.path.basename(file_path)
    if file_path.endswith('.ipynb'):
        logging.info('Convert notebook to python script')
        target_file = convert_notebook(file_path)
        file_path = target_file
        file_name = os.path.basename(file_path)
    else:
        # write edited code to different file
        target_file = os.path.join(os.path.dirname(file_path), 'component_' + file_name)

    target_file_name = os.path.basename(target_file)

    with open(file_path, 'r') as f:
        script = f.read()
    # Add code for logging and cli parameters to the beginning of the script
    script = gw_component_setup_code + script
    # replace old filename with new file name
    script = script.replace(file_name, target_file_name)
    with open(target_file, 'w') as f:
        f.write(script)

    if '__main__' not in script:
        logging.warning('No __main__ found in component code. Grid wrapper will import functions from component, '
                        'which can lead to unexpected behaviour without using __main__.')

    logging.info('Saved component python script in ' + target_file)

    return target_file


def apply_grid_wrapper(file_path, component_process, cos, *args, **kwargs):
    assert file_path.endswith('.py') or file_path.endswith('.ipynb'), \
        "Please provide a component file path to a python script or notebook."

    file_path = edit_component_code(file_path)

    description, interface, inputs, dependencies = get_component_elements(file_path)

    component_elements = dict(
        component_path=file_path,
        component_description=description,
        component_dependencies=dependencies,
        component_interface=interface,
        component_inputs=inputs,
        component_process=component_process
    )

    logging.debug('Wrap component with parameters:')
    for component, value in component_elements.items():
        logging.debug(component + ':\n' + str(value) + '\n')

    logging.info('Wrap component')
    grid_wrapper_file_path = wrap_component(cos=cos, **component_elements)
    return grid_wrapper_file_path, file_path


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('file_path', type=str,
                        help='Path to python script or notebook')
    parser.add_argument('additional_files', type=str, nargs='*',
                        help='List of paths to additional files to include in the container image')
    parser.add_argument('-p', '--component_process', type=str, required=True,
                        help='Name of the component sub process that is executed for each batch.')
    parser.add_argument('--cos', action=argparse.BooleanOptionalAction, default=False,
                        help='Creates a grid wrapper for processing COS files')
    parser.add_argument('-r', '--repository', type=str, default=None,
                        help='Container registry address, e.g. docker.io/<your_username>')
    parser.add_argument('-v', '--version', type=str, default=None,
                        help='Image version')
    parser.add_argument('-l', '--log_level', type=str, default='INFO')
    parser.add_argument("-v", "--version", action="store_true")
    parser.add_argument('--dockerfile_template_path', type=str, default='',
                        help='Path to custom dockerfile template')

    args = parser.parse_args()

    if args.version:
        print(version("claimed-c3"))
        sys.exit()

    # Init logging
    root = logging.getLogger()
    root.setLevel(args.log_level)
    handler = logging.StreamHandler(sys.stdout)
    formatter = logging.Formatter('%(levelname)s - %(message)s')
    handler.setFormatter(formatter)
    handler.setLevel(args.log_level)
    root.addHandler(handler)

    grid_wrapper_file_path, component_path = apply_grid_wrapper(**vars(args))

    if args.repository is not None:
        logging.info('Generate CLAIMED operator for grid wrapper')

        # Add component path and init file path to additional_files
        args.additional_files.append(component_path)

        # Update dockerfile template if specified
        if args.dockerfile_template_path != '':
            logging.info(f'Uses custom dockerfile template from {args.dockerfile_template_path}')
            with open(args.dockerfile_template_path, 'r') as f:
                dockerfile_template = Template(f.read())

        create_operator(
            file_path=grid_wrapper_file_path,
            repository=args.repository,
            version=args.version,
            dockerfile_template=dockerfile_template,
            additional_files=args.additional_files,
            log_level=args.log_level,
        )

        logging.info('Remove local component file')
        os.remove(component_path)


if __name__ == '__main__':
    main()



import json
import logging
import os
import re
from c3.parser import ContentParser


class Pythonscript:
    def __init__(self, path, function_name: str = None):

        self.path = path
        with open(path, 'r') as f:
            self.script = f.read()

        self.name = os.path.basename(path)[:-3].replace('_', '-')
        assert '"""' in self.script, 'Please provide a description of the operator in the first doc string.'
        self.description = self.script.split('"""')[1].strip()
        self.envs = self._get_env_vars()

    def _get_env_vars(self):
        cp = ContentParser()
        env_names = cp.parse(self.path)['env_vars']
        return_value = dict()
        for env_name in env_names:
            comment_line = str()
            for line in self.script.split('\n'):
                if re.search("[\"']" + env_name + "[\"']", line):
                    # Check the description for current variable
                    if not comment_line.strip().startswith('#'):
                        # previous line was no description, reset comment_line.
                        comment_line = ''
                    if comment_line == '':
                        logging.info(f'Interface: No description for variable {env_name} provided.')
                    if "int(" in line:
                        type = 'Integer'
                    elif "float(" in line:
                        type = 'Float'
                    elif "bool(" in line:
                        type = 'Boolean'
                    else:
                        type = 'String'
                    if ',' in line:
                        default = line.split(',', 1)[1].rstrip(') ').strip().replace("\"", "\'")
                    else:
                        default = None
                    return_value[env_name] = {
                        'description': comment_line.replace('#', '').replace("\"", "\'").strip(),
                        'type': type,
                        'default': default
                    }
                    break
                comment_line = line
        return return_value

    def get_requirements(self):
        requirements = []
        for line in self.script.split('\n'):
            pattern = r"([ ]*pip[ ]*install[ ]*)([A-Za-z=0-9.\-: ]*)"
            result = re.findall(pattern, line)
            if len(result) == 1:
                requirements.append((result[0][0].strip() + ' ' + result[0][1].strip()))
        return requirements

    def get_name(self):
        return self.name

    def get_description(self):
        return self.description

    def get_inputs(self):
        return {key: value for (key, value) in self.envs.items() if not key.startswith('output_')}

    def get_outputs(self):
        return {key: value for (key, value) in self.envs.items() if key.startswith('output_')}



# TODO: Rename the file to the desired operator name.
"""
TODO: Update the description of the operator in the first doc string.
This is the operator description.
The file name becomes the operator name.
"""

# TODO: Update the required pip packages.
# pip install numpy

import os
import logging
import numpy as np

# TODO: Add the operator interface.
# A comment one line above os.getenv is the description of this variable.
input_path = os.getenv('input_path')

# Optionally, you can set a default values.
with_default = os.getenv('with_default', 'default_value')

# You can cast a specific type with int(), float(), or bool().
num_values = int(os.getenv('num_values', 5))

# Output paths are starting with "output_".
output_path = os.getenv('output_path', None)


# You can call a function from an additional file (must be in the same directory) or add your code here.
def main(num_values, *args, **kwargs):
    # TODO: Add your code.
    random_values = np.random.rand(num_values)
    # C3 adds setup code to your script which initalize the logging.
    # You can just use logging.debug(), logging.info(), logging.warning() in your code.
    logging.info(f'Random values: {random_values}')


# It is recommended to use a main block to avoid unexpected code execution.
if __name__ == '__main__':
    main(num_values)


# TODO: Add a grid process if you want to parallelize your code.
def grid_process(batch_id, input_path, with_default, num_values, output_path):
    """
    A process for the c3 grid wrapper. The process gets the batch name as the first positional argument,
    followed by all interface variables. This is only possible if the code can be processed in parallel,
    e.g., by splitting up input files.
    """

    # You might need to update the variables based on the batch
    input_path += batch_id + '*.json'
    output_path += batch_id + '_data.csv'

    # Execute the processing with adjusted variables
    main(num_values, input_path, output_path)

    # optionally return a string or list with output files
    return output_path


"""
${component_name} got wrapped by grid_wrapper, which wraps any CLAIMED component and implements the generic grid computing pattern https://romeokienzler.medium.com/the-generic-grid-computing-pattern-transforms-any-sequential-workflow-step-into-a-transient-grid-c7f3ca7459c8

CLAIMED component description: ${component_description}
"""

# component dependencies
# ${component_dependencies}

import os
import json
import random
import logging
import time
import glob
from pathlib import Path

# import component code
from ${component_name} import *


# File with batches. Provided as a comma-separated list of strings or keys in a json dict.
gw_batch_file = os.environ.get('gw_batch_file', None)
# file path pattern like your/path/**/*.tif. Multiple patterns can be separated with commas. Is ignored if gw_batch_file is provided.
gw_file_path_pattern = os.environ.get('gw_file_path_pattern', None)
# pattern for grouping file paths into batches like ".split('.')[-1]". Is ignored if gw_batch_file is provided.
gw_group_by = os.environ.get('gw_group_by', None)
# path to grid wrapper coordinator directory
gw_coordinator_path = os.environ.get('gw_coordinator_path')
# lock file suffix
gw_lock_file_suffix = os.environ.get('gw_lock_file_suffix', '.lock')
# processed file suffix
gw_processed_file_suffix = os.environ.get('gw_lock_file_suffix', '.processed')
# error file suffix
gw_error_file_suffix = os.environ.get('gw_error_file_suffix', '.err')
# timeout in seconds to remove lock file from struggling job (default 1 hour)
gw_lock_timeout = int(os.environ.get('gw_lock_timeout', 3600))
# ignore error files and rerun batches with errors
gw_ignore_error_files = bool(os.environ.get('gw_ignore_error_files', False))

# component interface
${component_interface}

def load_batches_from_file(batch_file):
    if batch_file.endswith('.json'):
        # load batches from keys of a json file
        logging.info(f'Loading batches from json file: {batch_file}')
        with open(batch_file, 'r') as f:
            batch_dict = json.load(f)
        batches = batch_dict.keys()

    else:
        # Load batches from comma-separated txt file
        logging.info(f'Loading comma-separated batch strings from file: {batch_file}')
        with open(batch_file, 'r') as f:
            batch_string = f.read()
        batches = [b.strip() for b in batch_string.split(',')]

    logging.info(f'Loaded {len(batches)} batches')
    logging.debug(f'List of batches: {batches}')
    assert len(batches) > 0, f"batch_file {batch_file} has no batches."
    return batches


def identify_batches_from_pattern(file_path_patterns, group_by):
    logging.info(f'Start identifying files and batches')
    batches = set()
    all_files = []

    # Iterate over comma-separated paths
    for file_path_pattern in file_path_patterns.split(','):
        logging.info(f'Get file paths from pattern: {file_path_pattern}')
        files = glob.glob(file_path_pattern.strip())
        assert len(files) > 0, f"Found no files with file_path_pattern {file_path_pattern}."
        all_files.extend(files)

    # get batches by applying the group by function to all file paths
    for path_string in all_files:
        part = eval('str(path_string)' + group_by, {"group_by": group_by, "path_string": path_string})
        assert part != '', f'Could not extract batch with path_string {path_string} and group_by {group_by}'
        batches.add(part)

    logging.info(f'Identified {len(batches)} batches')
    logging.debug(f'List of batches: {batches}')

    return batches


def perform_process(process, batch):
    logging.debug(f'Check coordinator files for batch {batch}.')
    # init coordinator files
    lock_file = Path(gw_coordinator_path) / (batch + gw_lock_file_suffix)
    error_file = Path(gw_coordinator_path) / (batch + gw_error_file_suffix)
    processed_file = Path(gw_coordinator_path) / (batch + gw_processed_file_suffix)

    if lock_file.exists():
        # remove strugglers
        if lock_file.stat().st_mtime < time.time() - gw_lock_timeout:
            logging.debug(f'Lock file {lock_file} is expired.')
            lock_file.unlink()
        else:
            logging.debug(f'Batch {batch} is locked.')
            return

    if processed_file.exists():
        logging.debug(f'Batch {batch} is processed.')
        return

    if error_file.exists():
        if gw_ignore_error_files:
            logging.info(f'Ignoring previous error in batch {batch} and rerun.')
        else:
            logging.debug(f'Batch {batch} has error.')
            return

    logging.debug(f'Locking batch {batch}.')
    lock_file.parent.mkdir(parents=True, exist_ok=True)
    lock_file.touch()

    # processing files with custom process
    logging.info(f'Processing batch {batch}.')
    try:
        target_files = process(batch, ${component_inputs})
    except Exception as err:
        logging.error(f'{type(err).__name__} in batch {batch}: {err}')
        # Write error to file
        with open(error_file, 'w') as f:
            f.write(f"{type(err).__name__} in batch {batch}: {err}")
        lock_file.unlink()
        logging.error(f'Continue processing.')
        return

    # optional verify target files
    if target_files is not None:
        if isinstance(target_files, str):
            target_files = [target_files]
        for target_file in target_files:
            if not os.path.exists(target_file):
                logging.error(f'Target file {target_file} does not exist for batch {batch}.')
    else:
        logging.info(f'Cannot verify batch {batch} (target files not provided).')

    logging.info(f'Finished Batch {batch}.')
    processed_file.touch()

    # Remove lock file
    if lock_file.exists():
        lock_file.unlink()
    else:
        logging.warning(f'Lock file {lock_file} was removed by another process. '
                        f'Consider increasing gw_lock_timeout (currently {gw_lock_timeout}s) to repeated processing.')



def process_wrapper(sub_process):
    delay = random.randint(1, 60)
    logging.info(f'Staggering start, waiting for {delay} seconds')
    time.sleep(delay)

    # Init coordinator dir
    coordinator_dir = Path(gw_coordinator_path)
    coordinator_dir.mkdir(exist_ok=True, parents=True)

    # get batches
    if gw_batch_file is not None and os.path.isfile(gw_batch_file):
        batches = load_batches_from_file(gw_batch_file)
    elif gw_file_path_pattern is not None and gw_group_by is not None:
        batches = identify_batches_from_pattern(gw_file_path_pattern, gw_group_by)
    else:
        raise ValueError("Cannot identify batches. "
                         "Provide valid gw_batch_file or gw_file_path_pattern and gw_group_by.")

    # Iterate over all batches
    for batch in batches:
        perform_process(sub_process, batch)

    # Check and log status of batches
    processed_status = [(coordinator_dir / (batch + gw_processed_file_suffix)).exists() for batch in batches]
    lock_status = [(coordinator_dir / (batch + gw_lock_file_suffix)).exists() for batch in batches]
    error_status = [(coordinator_dir / (batch + gw_error_file_suffix)).exists() for batch in batches]

    logging.info(f'Finished current process. Status batches: '
                 f'{sum(processed_status)} processed / {sum(lock_status)} locked / {sum(error_status)} errors / {len(processed_status)} total')

    if sum(error_status):
        logging.error(f'Found errors! Resolve errors and rerun operator with gw_ignore_error_files=True.')
        # print all error messages
        for error_file in coordinator_dir.glob('**/*' + gw_error_file_suffix):
            with open(error_file, 'r') as f:
                logging.error(f.read())


if __name__ == '__main__':
    process_wrapper(${component_process})




import os
from string import Template
from pathlib import Path

# template file names
COMPONENT_SETUP_CODE = 'component_setup_code.py'
GW_COMPONENT_SETUP_CODE = 'gw_component_setup_code.py'
DOCKERFILE_FILE = 'dockerfile_template'
KFP_COMPONENT_FILE = 'kfp_component_template.yaml'
KUBERNETES_JOB_FILE = 'kubernetes_job_template.job.yaml'
GRID_WRAPPER_FILE = 'grid_wrapper_template.py'
COS_GRID_WRAPPER_FILE = 'cos_grid_wrapper_template.py'

# load templates
template_path = Path(os.path.dirname(__file__))

with open(template_path / COMPONENT_SETUP_CODE, 'r') as f:
    component_setup_code = f.read()

with open(template_path / GW_COMPONENT_SETUP_CODE, 'r') as f:
    gw_component_setup_code = f.read()

with open(template_path / DOCKERFILE_FILE, 'r') as f:
    dockerfile_template = Template(f.read())

with open(template_path / KFP_COMPONENT_FILE, 'r') as f:
    kfp_component_template = Template(f.read())

with open(template_path / KUBERNETES_JOB_FILE, 'r') as f:
    kubernetes_job_template = Template(f.read())

with open(template_path / GRID_WRAPPER_FILE, 'r') as f:
    grid_wrapper_template = Template(f.read())

with open(template_path / COS_GRID_WRAPPER_FILE, 'r') as f:
    cos_grid_wrapper_template = Template(f.read())



"""
This is the operator description.
"""

# pip install numpy

#!pip install pandas

import os
import numpy as np

# A comment one line above os.getenv is the description of this variable.
input_path = os.getenv('input_path')

# type casting to int(), float(), or bool()
batch_size = int(os.getenv('batch_size', 16))

# Commas in the previous comment are deleted because the yaml file requires descriptions without commas.
debug = bool(os.getenv('debug', False))

# Output parameters are starting with "output_"
output_path = os.getenv('output_path')


def main(*args):
    """
    The compiler only includes the first doc string.This text should not be included.
    """
    _ = np.random.randn(5)

    print(args)

def process(batch, *args):
    # process function for grid wrapper
    print('Execute batch:', batch)
    main(batch, *args, input_path, batch_size, debug, output_path)

if __name__ == '__main__':
    main(input_path, batch_size, debug, output_path)
